"""Tests file operations."""
# pylint: disable=redefined-outer-name

import json
import tempfile
import uuid as uuidlib
from datetime import datetime
from pathlib import Path

from responses import RequestsMock
from spectrumx import Client
from spectrumx.gateway import API_TARGET_VERSION
from spectrumx.ops.files import get_file_permissions

from tests.conftest import get_files_endpoint


def _download_file_endpoint(client: Client, file_id: str) -> str:
    """Returns the endpoint for downloading a file."""
    return (
        client.base_url + f"/api/{API_TARGET_VERSION}/assets/files/{file_id}/download/"
    )


def _get_file_id_endpoint(client: Client, file_id: str) -> str:
    """Returns the endpoint for getting a file."""
    return client.base_url + f"/api/{API_TARGET_VERSION}/assets/files/{file_id}/"


def _get_content_check_endpoint(client: Client) -> str:
    """Returns the endpoint for the content check API."""
    return (
        client.base_url
        + f"/api/{API_TARGET_VERSION}/assets/utils/check_contents_exist/"
    )


def test_get_file_permissions(temp_file_empty: Path) -> None:
    """Test get_file_permissions for many permission combinations."""
    chmod_combos = {
        "---------": 0o000,
        "--------x": 0o001,
        "r-xr-xr--": 0o554,
        "r--rw-r--": 0o464,
        "rw-rw-r--": 0o664,
        "rwx------": 0o700,
        "rwxrwxrwx": 0o777,
    }
    for perm_string, chmod in chmod_combos.items():
        temp_file_empty.chmod(chmod)
        assert get_file_permissions(temp_file_empty) == perm_string


def test_get_file_by_id(client: Client, responses: RequestsMock) -> None:
    """Given a file ID, the client must return the file."""
    uuid = uuidlib.uuid4()
    url: str = get_files_endpoint(client) + f"{uuid.hex}/"
    responses.get(
        url=url,
        status=200,
        json={
            "created_at": "2021-10-01T12:00:00Z",
            "directory": "/my/files/are/here/",
            "expiration_date": "2021-10-01T12:12:00Z",
            "media_type": "text/plain",
            "name": "file.txt",
            "permissions": "rw-rw-r--",
            "size": 321,
            "updated_at": "2021-10-01T12:00:00Z",
            "uuid": uuid.hex,
        },
    )
    client.dry_run = False  # to test the actual request
    file_sample = client.get_file(file_uuid=uuid.hex)
    assert file_sample.uuid == uuid


def test_file_get_returns_valid(
    client: Client,
) -> None:
    """The get_file method must return a valid File instance.

    Note the file may not exist locally, but the File instance can still be valid.
        That is the case for dry-runs - which use sample files - and for files fetched
        from the server that have not yet been downloaded.
    """

    file_id = uuidlib.uuid4()
    file_size = 888
    client.dry_run = True
    assert client.dry_run is True, "Dry run must be enabled for this test."
    file_sample = client.get_file(file_uuid=file_id.hex)
    assert file_sample.is_sample is True, "The file must be a sample file"  # pyright: ignore[reportPrivateUsage]
    assert file_sample.uuid is not None
    assert file_sample.name.startswith("dry-run-")
    assert file_sample.media_type == "text/plain"
    assert file_sample.size == file_size
    assert file_sample.directory == Path("sds-files/dry-run/")
    assert file_sample.permissions == "rw-rw-r--"
    assert isinstance(file_sample.created_at, datetime)
    assert isinstance(file_sample.updated_at, datetime)
    assert isinstance(file_sample.expiration_date, datetime), (
        "Expiration date should be a datetime"
    )
    assert file_sample.expiration_date > file_sample.created_at


def test_file_upload_returns_file(
    client: Client, temp_file_with_text_contents: Path
) -> None:
    """The upload_file method must return a valid File instance."""
    test_file_size = temp_file_with_text_contents.stat().st_size
    client.dry_run = True
    assert client.dry_run is True, "Dry run must be enabled for this test."
    file_sample = client.upload_file(
        local_file=temp_file_with_text_contents,
        sds_path=Path("/my/upload/location"),
    )
    assert file_sample.is_sample is False, (
        "The file must be a real file on disk (not a sample), even for this test"
    )  # pyright: ignore[reportPrivateUsage]
    assert temp_file_with_text_contents == file_sample.local_path, (
        "Local path does not match"
    )
    assert file_sample.uuid is None, "A local file must not have a UUID"
    assert file_sample.name is not None, "Expected a file name"
    assert file_sample.media_type == "text/plain", "Expected media type 'text/plain'"
    assert file_sample.size == test_file_size, "Expected the test file to be 4030 bytes"
    assert "/tmp/pytest-" in str(  # noqa: S108
        file_sample.local_path
    ), "Expected the temp file directory"
    assert file_sample.directory == Path("/my/upload/location")
    assert file_sample.permissions == "rw-r--r--"
    assert isinstance(file_sample.created_at, datetime)
    assert isinstance(file_sample.updated_at, datetime)
    assert file_sample.expiration_date is None, (
        "Local files should not have an expiration date"
    )


def test_large_file_upload_mocked(
    client: Client, responses: RequestsMock, temp_large_binary_file: Path
) -> None:
    """Test the upload_file method with mocked responses."""
    file_id = uuidlib.uuid4()
    file_size = temp_large_binary_file.stat().st_size
    client.dry_run = False  # calls are mocked, but we want to test the actual requests
    mocked_file_content_check_json = {
        "file_contents_exist_for_user": False,
        "file_exists_in_tree": False,
        "user_mutable_attributes_differ": True,
    }
    responses.add(
        method=responses.POST,
        url=_get_content_check_endpoint(client),
        status=201,
        json=mocked_file_content_check_json,
    )
    mocked_upload_json = {
        "uuid": file_id.hex,
        "name": temp_large_binary_file.name,
        "media_type": "text/plain",
        "size": file_size,
        "directory": "/my/custom/sds/location",
        "permissions": "rw-r--r--",
        "created_at": "2024-12-01T12:00:00Z",
        "updated_at": "2024-12-01T12:00:00Z",
        "expiration_date": "2026-12-01T12:00:00Z",
    }
    responses.add(
        method=responses.POST,
        url=get_files_endpoint(client),
        status=201,
        json=mocked_upload_json,
    )
    # run the test
    file_sample = client.upload_file(
        local_file=temp_large_binary_file,
        sds_path=Path(mocked_upload_json["directory"]),
    )
    assert file_sample.uuid == file_id, "UUID not as mocked."
    assert file_sample.name == mocked_upload_json["name"]
    assert file_sample.media_type == mocked_upload_json["media_type"]
    assert file_sample.size == file_size
    assert file_sample.directory == Path(mocked_upload_json["directory"])
    assert file_sample.permissions == mocked_upload_json["permissions"]
    assert isinstance(file_sample.created_at, datetime)
    assert isinstance(file_sample.updated_at, datetime)


def test_download_file_contents(client: Client, responses: RequestsMock) -> None:
    """The download_file_contents method must create a file with the file contents."""
    client.dry_run = False  # calls are mocked, but we want to test the actual requests

    # file properties
    file_id = uuidlib.uuid4()
    long_file_contents = (
        "File start\n" + "Sample file contents - " * 1000 + "\nFile end"
    )
    num_bytes = len(long_file_contents.encode("utf-8"))

    file_metadata = {
        "uuid": file_id.hex,
        "name": "downloaded-file.txt",
        "media_type": "text/plain",
        "size": num_bytes,
        "directory": "/my/download/location",
        "permissions": "rw-r--r--",
        "created_at": "2024-12-01T12:00:00Z",
        "updated_at": "2024-12-01T12:00:00Z",
        "expiration_date": "2026-12-01T12:00:00Z",
    }

    # mock the file metadata endpoint
    responses.add(
        method=responses.GET,
        url=_get_file_id_endpoint(client, file_id=file_id.hex),
        status=200,
        body=json.dumps(file_metadata),
    )

    # mock the file download endpoint
    responses.add(
        method=responses.GET,
        url=_download_file_endpoint(client, file_id=file_id.hex),
        status=200,
        body=long_file_contents,
    )

    # run the test
    downloaded_file = client.download_file(file_uuid=file_id.hex)
    downloaded_path = downloaded_file.local_path
    assert downloaded_path is not None, "Returned path must not be None"
    assert downloaded_path.exists(), "File was not downloaded to returned path"
    assert downloaded_path.is_file(), "Returned path must be a file"
    assert downloaded_path.stat().st_size == num_bytes, (
        f"File size must match. Got {num_bytes} B, "
        f"expected {downloaded_file.stat().st_size} B"
    )

    # cleanup
    downloaded_path.unlink(missing_ok=False)


def test_download_file_to_path(client: Client, responses: RequestsMock) -> None:
    """The download method must respect the given path, creating any parent dirs."""
    client.dry_run = False  # calls are mocked, but we want to test the actual requests

    # file properties
    file_id = uuidlib.uuid4()
    file_contents = "File start\n" + "Sample file contents" + "\nFile end"
    num_bytes = len(file_contents.encode("utf-8"))
    random_str = uuidlib.uuid4().hex
    parent_dir = Path(tempfile.gettempdir()) / random_str
    expected_path = parent_dir / "downloaded-file.txt"

    # mock the file metadata endpoint
    file_metadata = {
        "uuid": file_id.hex,
        "name": expected_path.name,
        "media_type": "text/plain",
        "size": num_bytes,
        "directory": "/my/download/location",
        "permissions": "rw-r--r--",
        "created_at": "2024-12-01T12:00:00Z",
        "updated_at": "2024-12-01T12:00:00Z",
        "expiration_date": "2026-12-01T12:00:00Z",
    }
    responses.add(
        method=responses.GET,
        url=_get_file_id_endpoint(client, file_id=file_id.hex),
        status=200,
        body=json.dumps(file_metadata),
    )

    # mock the file download endpoint
    responses.add(
        method=responses.GET,
        url=_download_file_endpoint(client, file_id=file_id.hex),
        status=200,
        body=file_contents,
    )

    # run the test
    assert parent_dir.exists() is False, "Parent dir must not exist before test"
    downloaded_file = client.download_file(
        file_uuid=file_id.hex, to_local_path=expected_path
    )
    downloaded_path = downloaded_file.local_path
    assert downloaded_path == expected_path, "Returned path must match the given path"
    assert expected_path.exists(), "File was not downloaded to given path"
    assert expected_path.is_file(), "Given path must be a file"
    assert expected_path.stat().st_size == num_bytes, (
        f"File size must match. Got {num_bytes} B, "
        f"expected {expected_path.stat().st_size} B"
    )

    # cleanup
    expected_path.unlink(missing_ok=False)
    parent_dir.rmdir()



"""Tests for SDS Captures."""

import json
from collections.abc import Generator
from pathlib import Path

import pytest
from loguru import logger as log
from spectrumx.client import Client
from spectrumx.models.captures import CaptureType
from spectrumx.utils import get_random_line

from tests.integration.conftest import PassthruEndpoints

log.trace("Placeholder log avoid reimporting or resolving unused import warnings.")

# paths with test data
dir_integration_data = Path(__file__).parent / "data"
dir_local_drf_top_level = dir_integration_data / "captures" / "drf" / "westford-vpol"
drf_channel = "cap-2024-06-27T14-00-00"


@pytest.fixture
def _capture_test() -> Generator[None]:
    """Choose to run capture test or not, based on existing data."""
    if not dir_local_drf_top_level.exists():
        pytest.skip(
            "Capture test SKIPPED. Test data does not "
            f"exist at: '{dir_local_drf_top_level}'"
        )

    # yield control to the test
    yield

    # teardown code for integration tests

    return


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_capture_test")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        [
            *PassthruEndpoints.file_content_checks(),
            *PassthruEndpoints.file_uploads(),
            *PassthruEndpoints.capture_creation(),
        ]
    ],
    indirect=True,
)
def test_capture_creation_drf(integration_client: Client) -> None:
    """Tests creating a Digital-RF capture."""

    # ARRANGE

    # metadata in this digital-rf capture in test data
    cap_start_bound: int = 1_719_499_740
    cap_is_continuous: bool = True
    cap_serial: str = "31649FE"

    # define paths in the context of a capture
    dir_top_level = dir_local_drf_top_level
    dir_channel = dir_top_level / drf_channel
    dir_metadata = dir_channel / "metadata"
    assert dir_metadata.is_dir(), (
        "Metadata directory should exist; check that "
        f"you have the right paths set: '{dir_metadata}'"
    )

    # suffix path_after_capture_data with a random name to avoid conflicts between runs
    path_after_capture_data = dir_top_level.relative_to(dir_integration_data)
    random_suffix = get_random_line(10, include_punctuation=False)
    path_after_capture_data = path_after_capture_data / f"test-{random_suffix}"

    _upload_assets(
        integration_client=integration_client,
        sds_path=path_after_capture_data,
        local_path=dir_top_level,
    )

    # ACT

    # create a capture
    capture = integration_client.captures.create(
        top_level_dir=Path(f"/{path_after_capture_data}"),
        channel=drf_channel,
        capture_type=CaptureType.DigitalRF,
    )

    # ASSERT

    # basic capture information
    assert capture.uuid is not None, "Capture UUID should not be None"
    assert capture.capture_type == CaptureType.DigitalRF
    assert capture.channel == drf_channel
    assert capture.top_level_dir == Path(f"/{path_after_capture_data}")

    # test capture properties
    assert capture.capture_props["start_bound"] == cap_start_bound
    assert capture.capture_props["is_continuous"] == cap_is_continuous
    assert (
        capture.capture_props["custom_attrs"]["receiver/info/mboard_serial"]
        == cap_serial
    )


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_capture_test")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        [
            *PassthruEndpoints.file_content_checks(),
            *PassthruEndpoints.file_uploads(),
            *PassthruEndpoints.capture_creation(),
        ]
    ],
    indirect=True,
)
def test_capture_creation_rh(integration_client: Client) -> None:
    """Tests creating a RadioHound capture."""

    # ARRANGE

    # define paths in the context of a capture
    dir_top_level = dir_integration_data / "captures" / "radiohound"
    radiohound_file = dir_top_level / "reference-v0.rh.json"
    assert radiohound_file.is_file(), (
        "Reference file should exist; check that "
        f"you have the right paths set: '{radiohound_file}'"
    )

    # suffix path_after_capture_data with a random name to avoid conflicts between runs
    rel_path_capture = dir_top_level.relative_to(dir_integration_data)
    random_suffix = get_random_line(10, include_punctuation=False)
    rel_path_capture = rel_path_capture / f"test-{random_suffix}"

    _upload_assets(
        integration_client=integration_client,
        sds_path=rel_path_capture,
        local_path=dir_top_level,
    )

    # ACT

    with radiohound_file.open("r") as fp_json:
        radiohound_data = json.load(fp_json)

    log.error(radiohound_data["scan_group"])

    # create a capture
    capture = integration_client.captures.create(
        top_level_dir=rel_path_capture,
        scan_group=radiohound_data["scan_group"],
        capture_type=CaptureType.RadioHound,
    )

    # ASSERT

    # basic capture information
    assert capture.uuid is not None, "Capture UUID should not be None"
    assert capture.capture_type == CaptureType.RadioHound
    assert capture.top_level_dir == rel_path_capture

    # test capture metadata
    assert capture.scan_group is not None, "Scan group should not be None"
    assert str(capture.scan_group) == radiohound_data["scan_group"], (
        "Scan group was not set correctly for RadioHound capture"
    )
    assert capture.capture_props, "Capture properties should not be empty"
    assert capture.capture_props == radiohound_data, (
        "Capture props doesn't match the reference data: \n"
        f"'{capture.capture_props}'\n!=\n'{radiohound_data}'"
    )


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_capture_test")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        [
            *PassthruEndpoints.capture_listing(),
        ]
    ],
    indirect=True,
)
def test_capture_listing_drf(integration_client: Client) -> None:
    """Tests reading and listing a Digital-RF capture."""
    captures = integration_client.captures.listing(
        capture_type=CaptureType.DigitalRF,
    )
    assert len(captures) > 0, "At least one capture should be present"
    for capture in captures:
        assert capture.uuid is not None, "Capture UUID should not be None"
        assert capture.capture_type == CaptureType.DigitalRF
        assert capture.channel is not None
        assert capture.top_level_dir is not None


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_capture_test")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        [
            *PassthruEndpoints.file_content_checks(),
            *PassthruEndpoints.file_uploads(),
            *PassthruEndpoints.capture_creation(),
        ]
    ],
    indirect=True,
)
def test_capture_update_rh(integration_client: Client) -> None:
    """Tests updating a RadioHound capture."""

    # ARRANGE

    # define paths in the context of a capture
    dir_top_level = dir_integration_data / "captures" / "radiohound"
    radiohound_file = dir_top_level / "reference-v0.rh.json"
    assert radiohound_file.is_file(), (
        "Reference file should exist; check that "
        f"you have the right paths set: '{radiohound_file}'"
    )

    # suffix path_after_capture_data with a random name to avoid conflicts between runs
    rh_capture_update_sds_path = dir_top_level.relative_to(dir_integration_data)
    random_suffix = get_random_line(10, include_punctuation=False)
    rh_capture_update_sds_path = rh_capture_update_sds_path / f"test-{random_suffix}"

    _upload_assets(
        integration_client=integration_client,
        sds_path=rh_capture_update_sds_path,
        local_path=dir_top_level,
    )

    with radiohound_file.open("r") as fp_json:
        radiohound_data = json.load(fp_json)

    # create a capture
    capture = integration_client.captures.create(
        top_level_dir=Path(f"/{rh_capture_update_sds_path}"),
        scan_group=radiohound_data["scan_group"],
        capture_type=CaptureType.RadioHound,
    )

    # ASSERT

    # basic capture information
    assert capture.uuid is not None, "Capture UUID should not be None"
    assert capture.capture_type == CaptureType.RadioHound
    assert capture.top_level_dir == Path(f"/{rh_capture_update_sds_path}")

    # test capture metadata
    assert capture.capture_props, "Capture properties should not be empty"
    assert capture.capture_props == radiohound_data, (
        "Capture props doesn't match the reference data: \n"
        f"'{capture.capture_props}'\n!=\n'{radiohound_data}'"
    )

    # upload a new radiohound file
    new_radiohound_file = dir_top_level / "reference-v0-addendum.rh.json"
    assert new_radiohound_file.is_file(), (
        "New reference file should exist; check that "
        f"you have the right paths set: '{new_radiohound_file}'"
    )

    _upload_assets(
        integration_client=integration_client,
        sds_path=rh_capture_update_sds_path,
        local_path=dir_top_level,
    )

    # ACT

    # update the capture
    integration_client.captures.update(
        capture_uuid=capture.uuid,
    )

    # ASSERT

    # if no exceptions occurred, the test passes
    assert True


def _upload_assets(
    integration_client: Client,
    sds_path: Path,
    local_path: Path,
) -> None:
    """Helper to upload a local directory to SDS and assert success."""
    log.info(f"Uploading assets as '/{sds_path}'")
    upload_results = integration_client.upload(
        local_path=local_path,
        sds_path=f"/{sds_path}",
        verbose=False,
    )
    success_results = [success for success in upload_results if success]
    failed_results = [success for success in upload_results if not success]
    assert len(failed_results) == 0, (
        f"No failed uploads should be present: {failed_results}"
    )
    log.debug(f"Uploaded {len(success_results)} assets.")



"""Capture model for SpectrumX."""

import sys

if sys.version_info < (3, 11):  # noqa: UP036
    from backports.strenum import StrEnum  # noqa: UP035 # Required backport
else:
    from enum import StrEnum

from pathlib import Path
from typing import Annotated
from typing import Any

from pydantic import UUID4
from pydantic import BaseModel
from pydantic import Field


class CaptureType(StrEnum):
    """Capture types in SDS."""

    DigitalRF = "drf"
    RadioHound = "rh"
    SigMF = "sigmf"


class CaptureOrigin(StrEnum):
    """Capture origins in SDS."""

    System = "system"
    User = "user"


_d_capture_props = "The indexed metadata for the capture"
_d_capture_type = f"The type of capture {', '.join([x.value for x in CaptureType])}"
_d_index_name = "The name of the SDS index associated with the capture"
_d_origin = "The origin of the capture"
_d_top_level_dir = "The top-level directory for the capture files"
_d_uuid = "The unique identifier for the capture"

_d_channel = "The channel associated with the capture. Only for RadioHound type."
_d_scan_group = "The scan group associated with the capture. Only for Digital-RF type."


class Capture(BaseModel):
    """A capture in SDS. A collection of spectrum data files that is indexed."""

    capture_props: Annotated[dict[str, Any], Field(description=_d_capture_props)]
    capture_type: Annotated[CaptureType, Field(description=_d_capture_type)]
    index_name: Annotated[str, Field(max_length=255, description=_d_index_name)]
    origin: Annotated[CaptureOrigin, Field(description=_d_origin)]
    top_level_dir: Annotated[Path, Field(description=_d_top_level_dir)]
    uuid: Annotated[UUID4, Field(description=_d_uuid)]

    # optional fields
    channel: Annotated[
        str | None, Field(max_length=255, description=_d_channel, default=None)
    ]
    scan_group: Annotated[UUID4 | None, Field(description=_d_scan_group, default=None)]

    def __str__(self) -> str:
        return f"<{self.__repr_name__} {self.uuid}>"


__all__ = [
    "Capture",
    "CaptureType",
]



"""Client for the SpectrumX Data System."""

from collections.abc import Mapping
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

from loguru import logger as log
from pydantic import UUID4

from spectrumx.api.captures import CaptureAPI
from spectrumx.errors import Result
from spectrumx.errors import SDSError
from spectrumx.ops.pagination import Paginator

from . import __version__
from .config import SDSConfig
from .gateway import GatewayClient
from .models.files import File
from .ops import files
from .utils import clean_local_path
from .utils import get_prog_bar
from .utils import log_user
from .utils import log_user_error
from .utils import log_user_warning

if TYPE_CHECKING:
    from tqdm import tqdm


class Client:
    """Instantiates and SDS client."""

    host: str
    is_authenticated: bool
    verbose: bool = False
    captures: CaptureAPI

    _gateway: GatewayClient
    _config: SDSConfig

    def __init__(
        self,
        host: None | str,
        *,
        env_file: Path | None = None,
        env_config: Mapping[str, Any] | None = None,
        verbose: bool = False,
    ) -> None:
        # avoids circular import
        from spectrumx.api import sds_files as _sds_files

        self._sds_files = _sds_files

        self.host = host
        self.is_authenticated = False
        self.verbose = verbose
        self._config = SDSConfig(
            env_file=env_file,
            env_config=env_config,
            sds_host=host,
            verbose=self.verbose,
        )

        # initialize the gateway
        self._gateway = GatewayClient(
            host=self.host,
            api_key=self._config.api_key,
            timeout=self._config.timeout,
            verbose=self.verbose,
        )

        # create internal API instances
        self.captures = CaptureAPI(
            gateway=self._gateway,
            dry_run=self.dry_run,
        )

        self._issue_user_alerts()

    def __str__(self) -> str:
        return f"Client(host={self.host})"

    def _issue_user_alerts(self) -> None:
        """Logs important messages on initialization."""
        if __version__.startswith("0.1."):
            log_user_warning(
                "This version of the SDK is in early development. "
                "Expect breaking changes in the future."
            )
        if self.dry_run:
            log_user("Dry run enabled: no SDS requests will be made or files written")

    @property
    def dry_run(self) -> bool:
        """When in dry run mode, no SDS requests are made and files are not written."""
        return self._config.dry_run

    @dry_run.setter
    def dry_run(self, value: bool) -> None:
        """Sets the dry run mode for the client and internal API instances."""

        self._config.dry_run = bool(value)
        self.captures.dry_run = self._config.dry_run
        # add future API instances here

        msg = (
            "Dry-run enabled: no SDS requests will be made or files written."
            if self._config.dry_run
            else "Dry-run DISABLED: modifications are now possible."
        )
        log_user_warning(msg)

    @property
    def base_url(self) -> str:
        """Base URL for the client."""
        return self._gateway.base_url

    @property
    def base_url_no_port(self) -> str:
        """Base URL without the port."""
        return self._gateway.base_url_no_port

    def authenticate(self) -> None:
        """Authenticate the client."""
        if self.dry_run:
            log_user("Dry run is enabled: assuming successful authentication")
            log_user("To authenticate against SDS, set Client.dry_run to False")
        else:
            log.warning(f"Dry run DISABLED: authenticating against '{self.host}'")
            self._gateway.authenticate()
        log.info("Authenticated successfully")
        self.is_authenticated = True

    # ======= FILE METHODS

    def get_file(self, file_uuid: UUID4 | str) -> File:
        """Get a file instance by its ID. Only metadata is downloaded from SDS.

        Note this does not download the file contents from the server. File
            instances still need to have their contents downloaded to create
            a local copy - see `Client.download_file()`.

        Args:
            file_uuid: The UUID of the file to retrieve.
        Returns:
            The file instance, or a sample file if in dry run mode.
        """

        return self._sds_files.get_file(client=self, file_uuid=file_uuid)

    def download(
        self,
        *,
        from_sds_path: Path | str,
        to_local_path: Path | str,
        skip_contents: bool = False,
        overwrite: bool = False,
        verbose: bool = True,
    ) -> list[Result[File]]:
        """Downloads files from SDS.

        Args:
            from_sds_path:  The virtual directory on SDS to download files from.
            to_local_path:  The local path to save the downloaded files to.
            skip_contents:  When True, only the metadata is downloaded.
            overwrite:      Whether to overwrite existing local files.
            verbose:        Show a progress bar.
        Returns:
            A list of results for each file discovered and downloaded.
        """
        from_sds_path = Path(from_sds_path)
        to_local_path = Path(to_local_path)

        # local vars
        prefix = "Dry-run: simulating download:" if self.dry_run else "Downloading:"

        if not to_local_path.exists():
            if self.dry_run:
                log_user(f"Dry run: would create the directory '{to_local_path}'")
            else:
                to_local_path.mkdir(parents=True)

        if self.dry_run:
            log_user(
                "Called download() in dry run mode: "
                "files are fabricated and not written to disk."
            )
            files_to_download = files.generate_random_files(num_files=10)
            log_user(f"Dry run: discovered {len(files_to_download)} files (samples)")
        else:
            files_to_download = self.list_files(sds_path=from_sds_path)
            if verbose:
                log_user(f"Discovered {len(files_to_download)} files")

        prog_bar: tqdm[File] = get_prog_bar(
            files_to_download, desc="Downloading", disable=not verbose
        )

        results: list[Result[File]] = []
        for file_info in prog_bar:
            prog_bar.set_description(f"{prefix} '{file_info.name}'")
            sds_dir = file_info.directory

            local_file_path = Path(f"{to_local_path}/{sds_dir}") / file_info.name
            local_file_path = clean_local_path(local_file_path)
            # this path will be validated later

            # skip download of local files (without UUID)
            if file_info.uuid is None:
                msg = f"Skipping local file: {file_info.name}"
                log_user_warning(msg)
                results.append(
                    Result(
                        exception=SDSError(msg),
                        error_info={
                            "file": file_info,
                        },
                    )
                )
                continue

            # register failure if the resolved path is not relative to the target
            local_file_path = local_file_path.resolve()
            to_local_path = to_local_path.resolve()
            log.debug(f"Resolved path: {local_file_path}")
            log.debug(f"Target path: {to_local_path}")
            if not local_file_path.is_relative_to(to_local_path):
                msg = (
                    f"Resolved path {local_file_path} is not relative to "
                    f"{to_local_path}: skipping download."
                )
                log_user_warning(msg)
                results.append(
                    Result(
                        exception=SDSError(msg),
                        error_info={
                            "file": file_info,
                        },
                    )
                )
                continue

            # avoid unintended overwrites (success)
            if local_file_path.exists() and not overwrite:
                log_user(f"Skipping existing file: '{local_file_path}'")
                results.append(Result(value=file_info))
                continue

            # download the file and register result
            try:
                log.debug(f"Dw: {local_file_path}")
                downloaded_file = self.download_file(
                    file_uuid=file_info.uuid,
                    to_local_path=local_file_path,
                    skip_contents=skip_contents,
                )
                results.append(Result(value=downloaded_file))
            except SDSError as err:
                log_user_error(f"Download failed: {err}")
                results.append(
                    Result(
                        exception=err,
                        error_info={
                            "file": file_info,
                        },
                    )
                )

        return results

    def list_files(
        self, sds_path: Path | str, *, verbose: bool = False
    ) -> Paginator[File]:
        """Lists files in a given SDS path.

        Args:
            sds_path: The virtual directory on SDS to list files from.
            verbose:  Show network requests and other info.
        Returns:
            A paginator for the files in the given SDS path.
        """
        return self._sds_files.list_files(
            client=self, sds_path=sds_path, verbose=verbose
        )

    def download_file(
        self,
        *,
        file_instance: File | None = None,
        file_uuid: UUID4 | str | None = None,
        to_local_path: Path | str | None = None,
        skip_contents: bool = False,
        warn_missing_path: bool = True,
    ) -> File:
        """Downloads a file from SDS: metadata and maybe contents.

        > Note this function always overwrites the local file, if it exists.

        Either `file_instance` or `file_uuid` must be provided. When passing
        a `file_instance`, it's recommended to set its `local_path` attribute,
        otherwise the download will create a temporary file on disk.

        Args:
            file_instance:      The file instance to download.
            file_uuid:          The UUID of the file to download.
            to_local_path:      The local path to save the downloaded file to.
            skip_contents:      When True, only the metadata is downloaded
                                    and no files are created on disk.
            warn_missing_path:  Show a warning when the download location is undefined.
        Returns:
            The file instance with updated attributes, or a sample when in dry run.
        """
        return self._sds_files.download_file(
            client=self,
            file_instance=file_instance,
            file_uuid=file_uuid,
            to_local_path=to_local_path,
            skip_contents=skip_contents,
            warn_missing_path=warn_missing_path,
        )

    def upload(
        self,
        *,
        local_path: Path | str,
        sds_path: Path | str = "/",
        verbose: bool = True,
    ) -> list[Result[File]]:
        """Uploads a file or directory to SDS.

        Args:
            local_path: The local path of the file or directory to upload.
            sds_path:   The virtual directory on SDS to upload the file to, \
                            where '/' is the user root.
            verbose:    Show a progress bar.
        """
        local_path = Path(local_path) if isinstance(local_path, str) else local_path
        valid_files = files.get_valid_files(local_path, warn_skipped=True)
        prog_bar = get_prog_bar(valid_files, desc="Uploading", disable=not verbose)
        upload_results: list[Result[File]] = []
        for file_path in prog_bar:
            try:
                result = Result(
                    value=self._sds_files.upload_file(
                        client=self,
                        local_file=file_path,
                        sds_path=sds_path,
                    )
                )
            except SDSError as err:
                log_user_error(f"Upload failed: {err}")
                result = Result(exception=err)
            upload_results.append(result)
        return upload_results

    def upload_file(
        self,
        *,
        local_file: File | Path | str,
        sds_path: Path | str = "/",
    ) -> File:
        """Uploads a file to SDS.

        If the file instance passed already has a directory set, `sds_path` will
        be prepended. E.g. given:
            `sds_path = 'baz'`; and
            `local_file.directory = 'foo/bar/'`, then:
        The file will be uploaded to `baz/foo/bar/` (under the user root in SDS) and
            the returned file instance will have its `directory` attribute updated
            to match this new path.

        Args:
            local_file:     The local file to upload.
            sds_path:       The virtual directory on SDS to upload the file to.
        Returns:
            The file instance with updated attributes, or a sample when in dry run.
        """

        return self._sds_files.upload_file(
            client=self,
            local_file=local_file,
            sds_path=sds_path,
        )


__all__ = ["Client"]



"""Smoke-like checks for build acceptance, including basic usage of the SDK.

Do not run with pytest. This file doubles as documentation, so it doesn't
    have pytest constructs and has some code repetition between examples
    for learning purposes.

The goal is to be a quick way to check if client code runs, without strict
    assertions or coverage. As such, I'm calling these "checks" instead of
    "tests".

Run it as a standalone script between the package build and publishing steps.
Avoid adding third-party imports to this file.
"""
# ruff: noqa: ERA001, T201, I001

import datetime
from collections.abc import Callable
from dataclasses import dataclass
from pathlib import Path
from random import randint
from random import random
from typing import TYPE_CHECKING

from spectrumx import Client
from spectrumx import enable_logging
from spectrumx.models.captures import Capture

NOT_IMPLEMENTED = "This example is not yet implemented."

if TYPE_CHECKING:
    from spectrumx.models.files import File
    from spectrumx.models.datasets import Dataset
    from spectrumx.ops.pagination import Paginator


def setup_module() -> None:
    """Setup any state specific to the execution of the given module."""
    Path("my_spectrum_files").mkdir(parents=True, exist_ok=True)


def teardown_module() -> None:
    """Teardown any state that was previously setup with a setup_module method."""
    Path("my_spectrum_files").rmdir()


def check_basic_usage() -> None:
    """Runs a basic usage run. Update this code to reflect readme changes."""

    # Example of files upload, listing, and download from SDS.

    # NOTE: the SDS client-server interaction is stateless, so it is
    #   not recommended to have multiple clients writing to the same
    #   locations simultaneously, as they may overrule each other
    #   and cause loss of data.
    sds = Client(
        host="sds.crc.nd.edu",
        # env_file=Path(".env"),  # default
        # env_config={"SDS_SECRET_TOKEN": "my-custom-token"},  # overrides
    )

    # when in dry-run (default), no changes are made to the SDS or the local filesystem
    # to enable the changes, set dry_run to False, as in:
    # sds.dry_run = False

    # authenticate using either the token from
    # the .env file or in the config passed in
    sds.authenticate()

    # local_dir has your own local files that will be uploaded to the SDS
    reference_name: str = "my_spectrum_files"
    local_dir: Path = Path(reference_name)

    # or, if the directory doesn't exist, let's create some fake data
    if not local_dir.exists():
        local_dir.mkdir(exist_ok=True)
        num_files = 10
        for file_idx in range(num_files):
            num_lines = randint(10, 100)  # noqa: S311
            file_name = f"rf_run_{file_idx}.csv"
            with (local_dir / file_name).open(mode="w", encoding="utf-8") as file_ptr:
                fake_nums = [random() for _ in range(num_lines)]  # noqa: S311
                file_ptr.write("\n".join(map(str, fake_nums)))

    # upload all files in a directory to the SDS
    # sds.dry_run = False   # uncomment to actually upload the files
    sds.upload(
        local_path=local_dir,  # may be a single file or a directory
        sds_path=reference_name,  # files will be created under this virtual directory
        verbose=True,  # shows a progress bar (default)
    )

    # download the files from an SDS directory
    # sds.dry_run = False
    local_downloads = Path("sds-downloads") / "files" / reference_name
    sds.download(
        from_sds_path=reference_name,  # files will be downloaded from this virtual dir
        to_local_path=local_downloads,  # download to this location (it may be created)
        overwrite=False,  # do not overwrite local existing files (default)
        verbose=True,  # shows a progress bar (default)
    )

    if not sds.dry_run:
        print("Downloaded files:")
        for file_path in local_downloads.iterdir():
            print(file_path)
    else:
        print("Turn off dry-run to download and write files.")


def check_error_handling() -> None:
    """Basic error handling examples."""

    # ======== Authentication ========

    from spectrumx.errors import AuthError, NetworkError

    sds = Client(host="sds.crc.nd.edu")
    try:
        sds.authenticate()
    except NetworkError as err:
        print(f"Failed to connect to the SDS: {err}")
        # check your host= parameter and network connection
        # if you're hosting the SDS Gateway, make sure it is accessible
    except AuthError as err:
        print(f"Failed to authenticate: {err}")
        # TODO: take action

    # ======== File operations ========

    from time import sleep
    from spectrumx.errors import NetworkError, SDSError, ServiceError

    # ...
    local_dir: Path = Path("my_spectrum_files")
    reference_name: str = "my_spectrum_files"
    is_success = False
    retries_left: int = 5
    while not is_success and retries_left > 0:
        try:
            retries_left -= 1
            # the sds.upload will restart a partial file transfer from zero,
            # but it won't re-upload already finished files.
            sds.upload(
                local_path=local_dir,
                sds_path=reference_name,
                verbose=True,
            )
            is_success = True
        except (NetworkError, ServiceError) as err:
            # NetworkError refers to connection issues between client and SDS Gateway
            # ServiceError refers to issues with the SDS Gateway itself (e.g. HTTP 500)
            # sleep longer with each retry, at least 5s, up to 5min
            sleep_time = max(5, 5 / (retries_left**2) * 60)
            print(f"Failed to reach the gateway; sleeping {sleep_time}s")
            print(f"Error: {err}")
            if retries_left > 0:
                sleep(sleep_time)
            continue
        except SDSError as err:
            print(f"Another SDS error occurred: {err}")
            # other errors might include e.g. OSError
            #   if listed files cannot be found.
            # TODO: take action or break
            break


def check_file_listing_usage() -> None:
    """Basic file listing usage example."""

    sds = Client(host="sds.crc.nd.edu")
    sds.authenticate()
    reference_name: str = "my_spectrum_files"

    # list files in an SDS directory, without downloading them
    files_generator = sds.list_files(sds_path=reference_name)

    # Note that the returned object is a (lazy) generator, and it is consumed after
    # the first iteration. This has some benefits:
    #   1. avoids loading all file metadata in memory at once;
    #   2. reduces the time for the first page of files; and
    #   3. avoids making more server requests than necessary, while spacing them out.

    # If you need to iterate over the files multiple times, you
    # can get the first N files from the generator into a list:
    num_files: int = 3
    up_to_three_files: list[File] = []
    for _ in range(num_files):
        try:
            up_to_three_files.append(next(files_generator))
        except StopIteration:
            break  # no more to list; we have less than `num_files` files
    # `up_to_three_files` may be iterated over again

    # Converting the entire paginator into a list is not recommended
    # for a large or unknown number of files, for the reasons above.

    # otherwise, we can just iterate over the generator
    print("Remaining files from the generator:")
    for file_entry in files_generator:
        print(f"\tProcessing {file_entry.name} of size {file_entry.size} B...")
        # do_something_with_the_file(file_entry)


def check_capture_usage() -> None:
    """Basic capture usage example."""
    raise NotImplementedError(NOT_IMPLEMENTED)

    sds = Client(host="sds.crc.nd.edu")
    sds.authenticate()
    local_downloads = Path("sds-downloads") / "captures"

    # search for capture files between two frequencies and dates
    # a "capture" represents a file in the SDS in a known format, such
    # as a Digital RF archive or SigMF file.
    utc = datetime.UTC
    start_time = datetime.datetime(2024, 1, 1, 0, 0, 0, tzinfo=utc)
    end_time = datetime.datetime(2024, 1, 2, 0, 0, 0, tzinfo=utc)

    # like in the file listing, the search results are a lazy-loaded Paginator
    captures: Paginator[Capture] = sds.search(
        asset_type=Capture,
        center_freq_range=(3e9, 5e9),  # between 3 and 5 GHz
        capture_time_range=(start_time, end_time),
        # additional arguments work as "and" filters
    )
    for capture in captures:
        print(f"Found capture: {capture.name} of type {capture.type}")
        # e.g. "Found capture: my_capture of type Digital RF"
        sds.captures.download(
            capture=capture,
            to_local_path=local_downloads / "search_results",
            overwrite=False,
            verbose=True,
        )


def check_dataset_usage() -> None:
    """Basic dataset usage example."""
    raise NotImplementedError(NOT_IMPLEMENTED)

    sds = Client(host="sds.crc.nd.edu")
    sds.authenticate()

    # get list of datasets available
    print("Dataset name | Dataset ID")
    for dataset in sds.datasets():
        print(f"{dataset.name} | {dataset.id}")

    local_downloads = Path("sds-downloads") / "datasets"

    # get the most recent dataset
    target_dataset: Dataset = sds.datasets.get_most_recent()

    # or fetch a dataset by its ID
    target_dataset: Dataset = sds.datasets.get(sds, dataset_id="dataset-id")

    # download a dataset to a local directory
    sds.download_dataset(
        dataset=target_dataset,
        to_local_path=local_downloads,  # download to this location + dataset_name
        overwrite=False,  # do not overwrite local files (default)
        verbose=True,  # shows a progress bar (default)
    )


@dataclass
class CheckCaller:
    call_fn: Callable[[], None]
    name: str

    def __call__(self) -> None:
        self.call_fn()


def log_header(msg: str) -> None:
    print(f"======= {msg.upper()}")


def main() -> None:
    """Runs all checks."""

    enable_logging()

    # check_basic_usage()
    all_checks = [
        CheckCaller(call_fn=check_basic_usage, name="Basic usage"),
        CheckCaller(call_fn=check_error_handling, name="Error handling"),
        CheckCaller(call_fn=check_file_listing_usage, name="File listing usage"),
        # uncomment the following when implemented:
        # CheckCaller(call_fn=check_capture_usage, name="Capture usage"),
        # CheckCaller(call_fn=check_dataset_usage, name="Dataset usage"),
    ]
    for check in all_checks:
        log_header(f"Running {check.name} check...".upper())
        check()
        log_header(f"{check.name} check completed.\n\n")


if __name__ == "__main__":
    main()



"""API functions specific to captures."""

from __future__ import annotations

import json
import uuid
from typing import TYPE_CHECKING
from uuid import uuid4

from loguru import logger as log

from spectrumx.models.captures import Capture
from spectrumx.models.captures import CaptureOrigin
from spectrumx.models.captures import CaptureType

if TYPE_CHECKING:
    from pathlib import Path

    from spectrumx.gateway import GatewayClient

index_mapping = {
    CaptureType.DigitalRF: "captures-drf",
    CaptureType.RadioHound: "captures-rh",
}


class CaptureAPI:
    client: GatewayClient

    def __init__(self, *, gateway: GatewayClient, dry_run: bool = True) -> None:
        """Initializes the CaptureAPI."""
        self.dry_run = dry_run
        self.gateway = gateway

    def create(
        self,
        *,
        top_level_dir: Path,
        capture_type: CaptureType,
        index_name: str = "",
        channel: str | None = None,
        scan_group: str | None = None,
    ) -> Capture:
        """Creates a new capture in SDS."""
        if index_name:
            log.warning(
                "The 'index_name' parameter is deprecated and "
                "will be removed in future versions."
            )
        index_name = index_mapping.get(capture_type, index_name)
        if not index_name:
            log.warning(f"Could not find an index for {capture_type=}")
        log.debug(
            f"Creating capture with {top_level_dir=}, "
            f"{channel=}, {capture_type=}, {index_name=}, {scan_group=}"
        )

        if self.dry_run:
            log.debug("Dry run enabled: simulating the capture creation")
            return Capture(
                capture_props={},
                capture_type=capture_type,
                channel=channel,
                index_name=index_name,
                origin=CaptureOrigin.User,
                scan_group=uuid.UUID(scan_group) if scan_group else None,
                top_level_dir=top_level_dir,
                uuid=uuid4(),
            )
        capture_raw = self.gateway.create_capture(
            capture_type=capture_type,
            channel=channel,
            index_name=index_name,
            scan_group=scan_group,
            top_level_dir=top_level_dir,
        )
        capture = Capture.model_validate_json(capture_raw)
        log.debug(f"Capture created with UUID {capture.uuid}")
        return capture

    def listing(self, *, capture_type: CaptureType) -> list[Capture]:
        """Lists all captures in SDS under the current user."""
        log.debug(f"Listing captures of type {capture_type}")
        captures_raw = self.gateway.list_captures(capture_type=capture_type)
        captures_list_raw = json.loads(captures_raw)
        captures: list[Capture] = []
        for captures_raw in captures_list_raw:
            capture = Capture.model_validate(captures_raw)
            captures.append(capture)
        log.debug(f"Listing {len(captures)} captures")
        return captures

    def update(
        self,
        *,
        capture_uuid: uuid.UUID,
    ) -> None:
        """Updates a capture in SDS by re-discovering and re-indexing the files.

        Re-upload the desired capture files in order to change
        """
        log.debug(f"Updating capture with UUID {capture_uuid}")

        if self.dry_run:
            log.debug(f"Dry run enabled: simulating capture update {capture_uuid}")
            return

        capture_raw = self.gateway.update_capture(
            capture_uuid=capture_uuid,
            verbose=True,
        )
        capture = Capture.model_validate_json(capture_raw)
        log.debug(f"Capture updated with UUID {capture.uuid}")

    def read(
        self,
        *,
        capture_uuid: uuid.UUID,
    ) -> Capture:
        """Reads a specific capture from SDS."""
        log.debug(f"Reading capture with UUID {capture_uuid}")

        capture_raw = self.gateway.read_capture(capture_uuid=capture_uuid)
        capture = Capture.model_validate_json(capture_raw)
        log.debug(f"Capture read with UUID {capture.uuid}")
        return capture
